// SPDX-License-Identifier: MIT
pragma solidity ^0.8.22;


// OpenZeppelin/utils/Context.sol
/**
 * @dev Provides information about the current execution context, including the
 * sender of the transaction and its data. While these are generally available
 * via msg.sender and msg.data, they should not be accessed in such a direct
 * manner, since when dealing with meta-transactions the account sending and
 * paying for execution may not be the actual sender (as far as an application
 * is concerned).
 *
 * This contract is only required for intermediate, library-like contracts.
 */
abstract contract Context {
    function _msgSender() internal view virtual returns (address) {
        return msg.sender;
    }

    function _msgData() internal view virtual returns (bytes calldata) {
        return msg.data;
    }
}

/**
 * @dev Contract module which provides a basic access control mechanism, where
 * there is an account (an owner) that can be granted exclusive access to
 * specific functions.
 *
 * By default, the owner account will be the one that deploys the contract. This
 * can later be changed with {transferOwnership}.
 *
 * This module is used through inheritance. It will make available the modifier
 * `onlyOwner`, which can be applied to your functions to restrict their use to
 * the owner.
 */
abstract contract Ownable is Context {
    address private _owner;

    event OwnershipTransferred(address indexed previousOwner, address indexed newOwner);

    /**
     * @dev Initializes the contract setting the deployer as the initial owner.
     */
    constructor() {
        _transferOwnership(_msgSender());
    }

    /**
     * @dev Returns the address of the current owner.
     */
    function owner() public view virtual returns (address) {
        return _owner;
    }

    /**
     * @dev Throws if called by any account other than the owner.
     */
    modifier onlyOwner() {
        require(owner() == _msgSender(), "Ownable: caller is not the owner");
        _;
    }

    /**
     * @dev Leaves the contract without owner. It will not be possible to call
     * `onlyOwner` functions anymore. Can only be called by the current owner.
     *
     * NOTE: Renouncing ownership will leave the contract without an owner,
     * thereby removing any functionality that is only available to the owner.
     */
    function renounceOwnership() public virtual onlyOwner {
        _transferOwnership(address(0));
    }

    /**
     * @dev Transfers ownership of the contract to a new account (`newOwner`).
     * Can only be called by the current owner.
     */
    function transferOwnership(address newOwner) public virtual onlyOwner {
        require(newOwner != address(0), "Ownable: new owner is the zero address");
        _transferOwnership(newOwner);
    }

    /**
     * @dev Transfers ownership of the contract to a new account (`newOwner`).
     * Internal function without access restriction.
     */
    function _transferOwnership(address newOwner) internal virtual {
        address oldOwner = _owner;
        _owner = newOwner;
        emit OwnershipTransferred(oldOwner, newOwner);
    }
}

contract ArrayCheck {

    uint256 internal BATCH_COUNT = 256;

    modifier checkArraySizeUint256(uint256[] memory x) {
        require(x.length > 0 && x.length <= BATCH_COUNT, "Too many elements in the array");
        _;
    }

    modifier checkArraySizeAddress(address[] memory x) {
        require(x.length > 0 && x.length <= BATCH_COUNT, "Too many elements in the array");
        _;
    }

    modifier checkArraySizeString(string[] memory x) {
        require(x.length > 0 && x.length <= BATCH_COUNT, "Too many elements in the array");
        _;
    }

    modifier checkArraySize1(uint256 len1) {
        require(len1 > 0 && len1 <= BATCH_COUNT, "The size of These Arrays must be equal");
        _;
    }

    modifier checkArraySize2(uint256 len1, uint256 len2) {
        require(len1 > 0 && len1 <= BATCH_COUNT && len1 == len2, "The size of These Arrays must be equal");
        _;
    }

    modifier checkArraySize3(uint256 len1, uint256 len2, uint256 len3) {
        require(len1 > 0 && len1 <= BATCH_COUNT && len1 == len2 && len2 == len3, "The size of These Arrays must be equal");
        _;
    }

    modifier checkArraySize4(uint256 len1, uint256 len2, uint256 len3, uint256 len4) {
        require(len1 > 0 && len1 <= BATCH_COUNT && len1 == len2 && len2 == len3 && len3 == len4, "The size of These Arrays must be equal");
        _;
    }

    modifier checkArraySize5(uint256 len1, uint256 len2, uint256 len3, uint256 len4, uint256 len5) {
        require(len1 > 0 && len1 <= BATCH_COUNT && len1 == len2 && len2 == len3 && len3 == len4 && len4 == len5, "The size of These Arrays must be equal");
        _;
    }

    modifier checkArraySize6(uint256 len1, uint256 len2, uint256 len3, uint256 len4, uint256 len5, uint256 len6) {
        require(len1 > 0 && len1 <= BATCH_COUNT && len1 == len2 && len2 == len3 && len3 == len4 && len4 == len5 && len5 == len6, "The size of These Arrays must be equal");
        _;
    }

    modifier checkArraySize7(uint256 len1, uint256 len2, uint256 len3, uint256 len4, uint256 len5, uint256 len6, uint256 len7) {
        require(len1 > 0 && len1 <= BATCH_COUNT && len1 == len2 && len2 == len3 && len3 == len4 && len4 == len5 && len5 == len6 && len6 == len7, "The size of These Arrays must be equal");
        _;
    }
}

contract Admin is Ownable, ArrayCheck {
    mapping(address => bool) private _admins;

    modifier onlyAdmin() {
        /*
        if (!_admins[msg.sender]) {
            console.log("onlyAdmin", address(this), msg.sender);
        }
        */
        require(_admins[msg.sender], "Only admin can call it");
        _;
    }

    constructor(){
        _admins[msg.sender] = true;
        // console.log("addAdmin", address(this), msg.sender);
    }

    function addAdmins(address[] memory admins_) public onlyOwner checkArraySizeAddress(admins_) {
        for (uint256 i = 0; i < admins_.length; i++) {
            addAdmin(admins_[i]);
        }
    }

    function delAdmins(address[] memory admins_) public onlyOwner checkArraySizeAddress(admins_) {
        for (uint256 i = 0; i < admins_.length; i++) {
            delAdmin(admins_[i]);
        }
    }

    function addAdmin(address admin_) public onlyOwner {
        _admins[admin_] = true;
        // console.log("addAdmin", address(this), _admin);
    }

    function delAdmin(address admin_) public onlyOwner {
        _admins[admin_] = false;
        // console.log("delAdmin", address(this), _admin);
    }

    function isAdmin(address _addr) public view returns (bool) {
        return _admins[_addr];
    }
}

contract Program is Admin {
//    using Address for address;
    mapping(address => bool) private _programs;

    modifier onlyProgram() {
        require(_programs[msg.sender], "Only program can call it");
        _;
    }

    function addPrograms(address[] memory programs_) public onlyOwner checkArraySizeAddress(programs_) {
        for (uint256 i = 0; i < programs_.length; i++) {
            addProgram(programs_[i]);
        }
    }

    function delPrograms(address[] memory programs_) public onlyOwner checkArraySizeAddress(programs_) {
        for (uint256 i = 0; i < programs_.length; i++) {
            delProgram(programs_[i]);
        }
    }

    function addProgram(address _program) public onlyOwner {
        //require(_program.isContract(), "Only contract can be a program");
        _programs[_program] = true;
    }

    function delProgram(address program_) public onlyOwner {
        _programs[program_] = false;
    }

    function isProgram(address program_) public view returns (bool) {
        return _programs[program_];
    }
}
